﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Aliyun.MNS.Model;
using Aliyun.MNS;
using System.Threading;
using System.Collections.Concurrent;

namespace ChangeMessageVisibilitySample
{
    class ChangeMessageVisibilitySample
    {
        private const string _accessKeyId = "XXX";
        private const string _secretAccessKey = "XXX";
        const string _endpoint = "XXX";

        // MaxParallelThreads set to 5 is enough in this sample
        const int MaxParallelThreads = 5;

        // Message should have at least MessageMinimalLife seconds life when given to WorkingMessageManager
        // To make this sample running fast, it is 5 seconds;
        const int MessageMinimalLife = 5;
        const int TimerInterval = 8;
        const int QueueMessageVisibilityTimeout = 30;

        // If message if not handled in certain time, we discard it
        // it could be 1 day or 12 hours
        const int MessageTimeout = 40;

        const string queueName = "TestMessageVisibilityQ";

        static IMNS sClient = new MNSClient(_accessKeyId, _secretAccessKey, _endpoint);

        static WorkingMessageManager sWorkingMessageManager = new WorkingMessageManager();

        // this manager now do only one thing
        //     change the message visibility to ensure that it is alive
        class WorkingMessageManager
        {
            protected Object lockObj = new Object();
            protected List<Message> workingBatch;
            protected List<Message> messages;
            protected ConcurrentDictionary<String, DateTime> messageRegisterTime;

            protected System.Timers.Timer updateTimer = new System.Timers.Timer();

            public WorkingMessageManager()
            {
                messageRegisterTime = new ConcurrentDictionary<string, DateTime>();
                workingBatch = new List<Message>();
                messages = new List<Message>();

                updateTimer.Interval = TimerInterval * 1000;
                updateTimer.AutoReset = true;
                updateTimer.Elapsed += OnUpdateTimerElapsed;
                updateTimer.Start();
            }

            protected void AddToMessages(Message message)
            {
                lock (lockObj)
                {
                    messages.Add(message);
                }
            }

            protected void SwitchMessagesAndWorking()
            {
                lock (lockObj)
                {
                    workingBatch = messages;
                    messages = new List<Message>();
                }
            }

            protected void OnUpdateTimerElapsed(object sender, System.Timers.ElapsedEventArgs args)
            {
                // switch working and later
                SwitchMessagesAndWorking();

                // now change visibility
                int count = workingBatch.Count;
                Parallel.For(0, count, new ParallelOptions { MaxDegreeOfParallelism = MaxParallelThreads },
                    i =>
                    {
                        ChangeMessageVisibility(workingBatch[i]);
                    });
            }

            protected void ChangeMessageVisibilityCallback(Object message)
            {
                ChangeMessageVisibility((Message)message);
            }

            protected void ChangeMessageVisibility(Message message)
            {
                var queue = sClient.GetNativeQueue(queueName);
                System.Diagnostics.Stopwatch stopWatch = System.Diagnostics.Stopwatch.StartNew();
                // although we have a timeout here, I think we could assume that it never timeout
                // and we believe that the request almost never fails.
                while (stopWatch.ElapsedMilliseconds < MessageMinimalLife * 1000)
                {
                    DateTime registerTime = new DateTime();
                    if (!messageRegisterTime.TryGetValue(message.Id, out registerTime))
                    {
                        Console.WriteLine("Message Removed! " + message.Id);
                        return;
                    }
                    else if (registerTime < DateTime.Now - new TimeSpan(0, 0, MessageTimeout))
                    {
                        Console.WriteLine("Message Timeout! " + message.Id);
                        messageRegisterTime.TryRemove(message.Id, out registerTime);
                        return;
                    }

                    if (message.NextVisibleTime
                        > DateTime.Now + new TimeSpan(0, 0, MessageMinimalLife + TimerInterval))
                    {
                        // do not need to change visibility
                        AddToMessages(message);
                        return;
                    }

                    try
                    {
                        ChangeMessageVisibilityResponse res = queue.ChangeMessageVisibility(message.ReceiptHandle, QueueMessageVisibilityTimeout);
                        message.ReceiptHandle = res.ReceiptHandle;
                        message.NextVisibleTime = res.NextVisibleTime;
                        AddToMessages(message);
                        Console.WriteLine("VisibilityChanged! " + message.Id);
                        return;
                    }
                    catch (MessageNotExistException me)
                    {
                        Console.WriteLine("ReceiptHandle Invalid! ");
                        break;
                    }
                    catch (Exception e)
                    {
                        System.Console.WriteLine("ChangeMessageVisibility Retry for: " + message.Id
                            + " " + message.ReceiptHandle + e.Message);
                        Thread.Sleep(1000);
                        continue;
                    }
                }

                // if it fails, we could inform the corresponding worker here (or start an error handling task)
                // BUT, it's better to execute "ErrorHandling" in another single thread
                Console.WriteLine("====================================================================");
                Console.WriteLine("ChangeMessageVisibility Failed for: " + message.Id);
                Console.WriteLine("====================================================================");
            }

            // workers should call registerMessage before handling the message
            // if registerMessage returns false, worker should throw the message away
            //     the message itself will come back soon
            public bool registerMessage(Message message)
            {
                if (message.Id == null
                    || message.ReceiptHandle == null
                    || message.NextVisibleTime <= DateTime.Now + new TimeSpan(0, 0, MessageMinimalLife)) {
                    return false;
                }

                messageRegisterTime.TryAdd(message.Id, DateTime.Now);
                ThreadPool.QueueUserWorkItem(new WaitCallback(ChangeMessageVisibilityCallback), message);
                return true;
            }

            public void deleteMessage(Message message)
            {
                try
                {
                    var queue = sClient.GetNativeQueue(queueName);
                    queue.DeleteMessage(message.ReceiptHandle);
                    Console.WriteLine("MessageDeleted");
                }
                catch (Exception e)
                {
                    Console.WriteLine("DeleteMessage Fails: " + e.Message);
                    // do something when message is not deleted
                }
                DateTime dateTime = new DateTime();
                messageRegisterTime.TryRemove(message.Id, out dateTime);
            }
        };

        public static void WorkerFunc(int workerId)
        {
            var queue = sClient.GetNativeQueue(queueName);
            Message message = null;
            try
            {
                ReceiveMessageResponse response = queue.ReceiveMessage();
                message = response.Message;
            }
            catch (Exception e)
            {
                Console.WriteLine("I am a cool worker. I NEVER RETRY!! " + e.Message);
                return;
            }

            if (!sWorkingMessageManager.registerMessage(message))
            {
                Console.WriteLine("THROW MESSAGE!");
                return;
            }
            try
            {
                // i just sleep without doing any work
                int count = QueueMessageVisibilityTimeout + 1;
                if (workerId == 2)
                {
                    count += MessageTimeout;
                }
                while (count > 0)
                {
                    count -= 3;
                    Console.WriteLine(workerId + ": I am working, not sleeping. " + message.Id);
                    Thread.Sleep(3000);
                }
            }
            finally
            {
                Console.WriteLine(workerId + ": I have done my work. " + message.Id);
                sWorkingMessageManager.deleteMessage(message);
            }
        }

        static void Main(string[] args)
        {
            try
            {
                CreateQueueRequest request = new CreateQueueRequest();
                request.QueueName = queueName;
                request.Attributes.VisibilityTimeout = QueueMessageVisibilityTimeout;

                sClient.CreateQueue(request);
                var queue = sClient.GetNativeQueue(queueName);

                var sendMessageRequest = new SendMessageRequest("阿里云<MessageBody>计算");
                queue.SendMessage(sendMessageRequest);
                queue.SendMessage(sendMessageRequest);
                queue.SendMessage(sendMessageRequest);
                queue.SendMessage(sendMessageRequest);
            }
            catch (Exception e)
            {
                Console.WriteLine("EXAMPLE FAIL: " + e.Message);
                return;
            }

            Parallel.For(0, 3, new ParallelOptions { MaxDegreeOfParallelism = 5 },
                    i =>
                    {
                        WorkerFunc(i);
                    });
        }
    }
}
